#ifndef QM_PLAYLIST_H
#define QM_PLAYLIST_H


////////////////////////////////////// PLAYLIST ///////////////////////////////////////
///////////////////// This is simply an extension of qm_Tracks.ui.h ///////////////////
////////////////////////////////////// TSILYALP ///////////////////////////////////////

void qm_Tracks::InitPlaylist()
{
    plistMax = 0;
    playlistSize = 0;
    ListTotalTime = 0;
    curSong = -1;
    b_waitForList  = TRUE;
    b_markplayed  = TRUE;
    b_waitForShuffle = FALSE;
    playListView->addColumn("#", -1);
    playListView->addColumn("", 16);
    playListView->addColumn("title", 0);
    playListView->addColumn("time", -1);
    playListView->setAllColumnsShowFocus(TRUE);
    playListView->setSortColumn(-1); // Disables sorting
    playListView->setSelectionMode(QListView::Extended);
    playListView->setResizeMode(QListView::NoColumn);
    playListView->setHScrollBarMode(QScrollView::AlwaysOff);
    playListView->setVScrollBarMode(QScrollView::AlwaysOn);
    playListView->setColumnAlignment (3, Qt::AlignRight);
    playListView->setType(LV_DRAGDROP);
    playListView->header()->setHidden(TRUE);

    QAccel *acc = new QAccel( this );
    acc->connectItem(acc->insertItem (Key_Delete), this, SLOT(plDeleteKeyPressed()));
    acc->connectItem(acc->insertItem (Key_Space), this, SLOT(spaceKeyPressed()));

    // colors & fonts: see qm_Tracks.h (init)

    // connections
    connect(parent(), SIGNAL(trSetCurSong(int,int)), this, SLOT(trSetCurSong(int,int)));
    connect(parent(), SIGNAL(trSetServerPlaylist(qm_SongPtrList)), this, SLOT(trSetServerPlaylist(qm_SongPtrList)));
    connect(parent(), SIGNAL(trGetsPlistInfo(int, const QString &, const QString &)), this, SLOT(trGetsPlistInfo(int, const QString &, const QString &)));
}

void qm_Tracks::keyPressEvent( QKeyEvent *e )
{
    if ( e->key() == Key_Control) // CTRL pressed
        b_CTRLpressed = TRUE;
}

void qm_Tracks::keyReleaseEvent( QKeyEvent *e )
{
    if ( e->key() == Key_Control) // CTRL released
        b_CTRLpressed = FALSE;
}

void qm_Tracks::trGetsPlistInfo(int plMax, const QString & plpath, const QString & muspath)
{
    plistMax = plMax;
    PlayListPath = plpath;
    MusicPath = muspath;
}


void qm_Tracks::playListClicked()
{
    dataBaseView->clearFocus();
    playListView->setFocus ();
}


void qm_Tracks::toggleMarkPlayed()
{
    b_markplayed = !b_markplayed;
    // Store in the Core
    crGetsTrData(dataMode, b_markplayed);
    if (!b_markplayed)
    {
    // hide the play-history tags
    qm_ListItem *findItem = (qm_ListItem*)(playListView->firstChild());
    while(findItem)
    {
        int stat = findItem->getItemStatus();
            if(findItem->getWasPlayed() == 1)
        {
            findItem->setWasPlayed(-1);
            if (stat != LITEM_STAT_CURRENT && stat != LITEM_STAT_STOPPED && stat != LITEM_STAT_PAUSED)
                findItem->setPixmap(1, trackIcon);
            findItem->repaint();
        }
            findItem = (qm_ListItem*)(findItem->itemBelow());
    }
    }
    else
    {
    // repaint the play-history tags
    qm_ListItem *findItem = (qm_ListItem*)(playListView->firstChild());
    while(findItem)
    {
        int stat = findItem->getItemStatus();
            if(findItem->getWasPlayed() == -1)
        {
            findItem->setWasPlayed(1);
            if (stat != LITEM_STAT_CURRENT && stat != LITEM_STAT_STOPPED && stat != LITEM_STAT_PAUSED)
                findItem->setPixmap(1, trackPlayedIcon);
            findItem->repaint();
        }
            findItem = (qm_ListItem*)(findItem->itemBelow());
    }
    }
}



void qm_Tracks::saveList()
{
    if(trConn == NULL)
    return;
    bool ok;
    QString text = QInputDialog::getText("Saving playlist", "Please enter a name for the list:", QLineEdit::Normal, QString::null, &ok, this );
    if ( !ok || text.isEmpty() )
    return;
    LockStatUpdate();
    mpd_sendSaveCommand(trConn, text.utf8());
    mpd_finishCommand(trConn);
    ErrorCheck(151,"mpd_sendSaveCommand");

    // special errorcheck (MPD_ERROR_ACK must not be ignored here)
    bool saveOK = TRUE;
    if(trConn == NULL)
    {
    // Signal the Core
    crConnectError();
    saveOK = FALSE;
    }
    else
    if(trConn->error)
    saveOK = FALSE;

    UnlockStatUpdate();

    if(!saveOK)
    QMessageBox::warning( this, "Error", "Could not save the list.\n" "Perhaps a list with this name already exists?" );
    else
    {
    // reload the playlist-list if it is shown
    if (dataMode == DMODE_LIST)
        trReloadData(FALSE);
    }
}


void qm_Tracks::trSetServerPlaylist(qm_SongPtrList theList)
{
    playlistSize = 0;
    ListTotalTime = 0;
    b_waitForList = TRUE;
    playListView->clear();
    qm_ListItem  *newItem;
    qm_ListItem  *prevItem = NULL;
    // loop through the list
    mpd_Song *aSong;
    for ( aSong = theList.first(); aSong; aSong = theList.next() )
    {
    mpd_Song thisSong = *aSong;
        // Skip if, for some reason, there is no file
    QString fileName = QString::fromUtf8(thisSong.file);
        if(fileName.isEmpty())
        continue;
    int songTime = thisSong.time;
    if (songTime < 0)
        songTime = 0;
        // The 1st to enter the loop has no prevItem
    if(prevItem)
    {
        newItem = new qm_ListItem(playListView, prevItem, thisSong, trackIcon, "#", "", sTitle(thisSong), formatTime(songTime) + " ");
        prevItem = newItem;
    }
    else
    {
        newItem = new qm_ListItem(playListView, thisSong, trackIcon, "#", "", sTitle(thisSong), formatTime(songTime) + " ");
        prevItem = newItem;
    }
    playlistSize++;
    //And make sure that they can be dragged :)
    newItem->setDragEnabled(TRUE);
    newItem->setType(LITEM_TRACK);
    ListTotalTime += songTime;
    }

    reNumberList();
    // no song playing
    curSong = -1;
    b_waitForList = FALSE;
    resetTime();
}

//  called when the user double clicks on a playlist item
void qm_Tracks::playThisOne(QListViewItem *theItem)
{
    if(trConn == NULL)
    return;
    // get the track nr. from the position
    QString pos = theItem->text(0);
    int nr = pos.toInt() - 1;
    LockStatUpdate();
    mpd_sendPlayCommand(trConn, nr);
    mpd_finishCommand(trConn);
    ErrorCheck(152,"Playlist mpd_sendPlayCommand");
    UnlockStatUpdate();
    // trigger update in Core:
    crCurSongChanged(-1);
}

// Called by the core. Marks the song that is playing
void qm_Tracks::trSetCurSong( int songNum, int status )
{
    // will be back here later
    if(b_waitForList || b_waitForShuffle)
    return;

    // nothing to do
    if (playlistSize == 0)
    return;

    QListViewItem  *findItem = 0;
    qm_ListItem *foundItem = 0;

    // curSong is about to be changed: reset current cursong.
    if(curSong != songNum)
    {
    printf("Track: " + QString::number(songNum) + " (was " + QString::number(curSong) + ")\n");

    if (curSong >= 0)
    {
            findPointer(curSong, playListView, findItem);
            foundItem = (qm_ListItem*)findItem;
            if (foundItem)
        {
            foundItem->setItemStatus(LITEM_STAT_NORMAL);
            if (b_markplayed && foundItem->getWasPlayed() == 1)
                foundItem->setPixmap(1, trackPlayedIcon);
            else
                foundItem->setPixmap(1, trackIcon);
            foundItem->repaint();
        }
    }
    }

    curSong = songNum;
    switch (status)
    {
    case MPD_STATUS_STATE_PAUSE:
    {
        findPointer(curSong, playListView, findItem);
        foundItem = (qm_ListItem*)findItem;
        if (foundItem)
        {
            foundItem->setPixmap(1, plPauseIcon);
            foundItem->setItemStatus(LITEM_STAT_PAUSED);
            foundItem->repaint();
        }
        break;
    }
    case MPD_STATUS_STATE_STOP:
    {
        findPointer(curSong, playListView, findItem);
        foundItem = (qm_ListItem*)findItem;
        if (foundItem)
        {
            foundItem->setPixmap(1, plStopIcon);
            foundItem->setItemStatus(LITEM_STAT_STOPPED);
            foundItem->repaint();
        }
        break;
    }
    case -1: // 'WAIT' (core sends this when no song is found)
    {
        findPointer(curSong, playListView, findItem);
        foundItem = (qm_ListItem*)findItem;
        if (foundItem)
        {
            foundItem->setPixmap(1, plWaitIcon);
            foundItem->setItemStatus(LITEM_STAT_STOPPED);
            foundItem->repaint();
        }
        break;
    }
    case MPD_STATUS_STATE_UNKNOWN: // reset
    {
        findPointer(curSong, playListView, findItem);
        foundItem = (qm_ListItem*)findItem;
        if (foundItem)
        {
            foundItem->setItemStatus(LITEM_STAT_NORMAL);
            foundItem->setPixmap(1, trackIcon);
            foundItem->repaint();
        }
        curSong = -1;
        break;
    }
    default:  // current song changed to new one which is playing, or resumed playing.
    {
        findPointer(songNum, playListView, findItem);
        foundItem = (qm_ListItem*)findItem;
        if (foundItem)
        {
            if (status == MPD_STATUS_STATE_PLAY) // should be true
            {
                if (b_markplayed)
                    foundItem->setWasPlayed(1);
                else
                    foundItem->setWasPlayed(-1);
            }
            foundItem->setItemStatus(LITEM_STAT_CURRENT);
            foundItem->setPixmap(1, plPlayIcon);
            foundItem->repaint();
        }
    }
    }
    // Possibly an error, but avoid endless feedback loops.
    if (!foundItem && songNum > 0 && playlistSize > 0)
    crUpdatePlaylist();
}

// Get the list item at position pos
void qm_Tracks::findPointer( const int pos, QListView *theTree, QListViewItem * & findItem )
{
    findItem = theTree->firstChild();
    for(int i = 0; i < pos; i++)
    {
        if(!findItem) return;
        findItem = findItem->itemBelow();
    }
}


// remove played tracks from the playlist
void qm_Tracks::purgePlayList()
{
    LockStatUpdate();
    QListViewItem *curItem;
    QListViewItem *ItemToDelete;
        bool curSongDeleted = FALSE;
    qm_CommandList newCommandList;
    // Start at the end!
    curItem = playListView->lastItem();
    while(curItem != 0)
    {
        char wasplayed = ((qm_ListItem*)curItem)->getWasPlayed();
        int  status = ((qm_ListItem*)curItem)->getItemStatus();
        if(wasplayed != 0 && status != LITEM_STAT_CURRENT &&  status != LITEM_STAT_PAUSED)
        {
            ItemToDelete = curItem;
            QString pos = ItemToDelete->text(0);
            int trackNumber = pos.toInt() - 1;
            if (trackNumber < curSong)
                curSong--;
            else
            if(trackNumber == curSong)
            {
                crSendStop();
                curSongDeleted = TRUE;
            }
            qm_CmdStruct newCommand;
            newCommand.cmd = QM_CMD_DEL;
            newCommand.trackNr = trackNumber;
            newCommand.time = ( ((qm_ListItem*)ItemToDelete)->getSong() ).time;
            if (newCommand.time < 0)
                newCommand.time = 0;
            newCommandList.append(newCommand);
            // move up the list
            curItem = curItem->itemAbove();
            playListView->takeItem(ItemToDelete);
            //And move on!
            continue;
            }
            // else just move one up
            curItem = curItem->itemAbove();
    }
    UnlockStatUpdate();

    executeList(newCommandList);

    if (curSongDeleted)
    {
        curSong = -1;
        if (playlistSize == 0)
            crUpdatePlaylist();
        else
            crCurSongChanged(-1);
    }
}


// Delete items from the playlist
void qm_Tracks::plDeleteKeyPressed()
{
    // Data-panel has the focus
    if (dataBaseView->hasFocus())
    {
    // Nothing to do if not in playlist-mode
    if (dataMode != DMODE_LIST)
        return;
    // Else check if any items are selected.
    QListViewItem *curItem;
    int count = 0;
    curItem = dataBaseView->firstChild()->firstChild(); // step over '/'
    while (curItem != NULL)
    {
        if (curItem->isSelected())
        {
            count++;
            break;
        }
        curItem = curItem->nextSibling();
    }
    if (count == 0)
        return;
    // Ask for confirmation
    if ( QMessageBox::question( this,tr("Delete playlists?"),
        tr("This will remove the selected playlists from your disk."), tr("&Yes"), tr("&No"), QString::null, 1, 0 ) )
        return;
    // Work to do!
    QListViewItem *ItemToDelete;
    curItem = dataBaseView->lastItem();
    LockStatUpdate();
    while(curItem != 0)
    {
        if(curItem->isSelected())
        {
            ItemToDelete = curItem;
            mpd_sendRmCommand(trConn, (ItemToDelete->text(0)).utf8());
            mpd_finishCommand(trConn);
            if(!ErrorCheck(101,"mpd_sendRmCommand"))
                break;

            curItem = curItem->itemAbove();
            dataBaseView->takeItem(ItemToDelete);
            continue;
        }
        // Move up
        curItem = curItem->itemAbove();
        }
        UnlockStatUpdate();
    trReloadData(FALSE); // this should not be necessary...
    }
    // removing item(s) from the playlist
    if (playListView->hasFocus())
    {
    LockStatUpdate();
    QListViewItem *curItem;
    QListViewItem *ItemToDelete;
        bool curSongDeleted = FALSE;
    qm_CommandList newCommandList;
    // Start at the end!
    curItem = playListView->lastItem();
    while(curItem != 0)
    {
        if(curItem->isSelected())
        {
            ItemToDelete = curItem;
            QString pos = ItemToDelete->text(0);
            int trackNumber = pos.toInt() - 1;
            if (trackNumber < curSong)
                curSong--;
            else
            if(trackNumber == curSong)
            {
                crSendStop();
                curSongDeleted = TRUE;
            }
            qm_CmdStruct newCommand;
            newCommand.cmd = QM_CMD_DEL;
            newCommand.trackNr = trackNumber;
            newCommand.time = ( ((qm_ListItem*)ItemToDelete)->getSong() ).time;
            if (newCommand.time < 0)
                newCommand.time = 0;
            newCommandList.append(newCommand);
            // move up the list
            curItem = curItem->itemAbove();
            playListView->takeItem(ItemToDelete);
            //And move on!
            continue;
            }
            // else just move one up
            curItem = curItem->itemAbove();

    }
    UnlockStatUpdate();

    executeList(newCommandList);

    if (curSongDeleted)
    {
        curSong = -1;
        if (playlistSize == 0)
            crUpdatePlaylist();
        else
            crCurSongChanged(-1);
    }
    }
}

// (Re)set the item numbers in the playlist
void qm_Tracks::reNumberList()
{
    // Get the first item
    QListViewItem *listItem = playListView->firstChild();
    int inc = 0;
    // Set track number and get next.
    while(listItem)
    {
        listItem->setText(0, " " + QString::number(++inc));
        listItem = listItem->itemBelow();
    }
    // Fix column widths (in case max went from ## to ### etc)
    plvResized();
    // make sure current song is the right one
    crCurSongChanged(-1);
}

// reset the text of the tracks-time label
void qm_Tracks::resetTime()
{
    lb_tracks_time->setText("Tracks: " + QString::number(playlistSize) + "   Total Time: " + listTime(ListTotalTime));
}


void qm_Tracks::addDraggedItems(qm_ListView *srcView, qm_ListItem *target)
{
    // dragging items to the playlist
    if (srcView == dataBaseView)
    {
    qm_CommandList commandList;
    qm_SongPtrList songList;
        // Iterate ALL (sub)items
    QListViewItemIterator Itr(srcView);
    while ( Itr.current() )
    {
        if ( Itr.current()->isSelected() )
        {
            qm_ListItem *thisItem = (qm_ListItem*)Itr.current();
            switch (thisItem->getType())
            {
                case LITEM_TRACK:
                {
                    songList.append( &(thisItem->getSong()) );
                    break;
                }
                case LITEM_DIR:
                {
                    addaDirectory((thisItem->getDir()).path, songList);
                    break;
                }
                case LITEM_ALBUM:
                {
                    addanAlbum(thisItem->getAlbum(), thisItem->getArtist(), songList);
                    break;
                }
                case LITEM_ARTIST:
                {
                    addanArtist(thisItem->getArtist(), songList);
                    break;
                }
                case LITEM_PLAYLIST:
                {
                    addaPlaylist(thisItem->getAlbum()); // list-name is stored in 'album'
                    break;
                }
                default:
                    break;
            }
            appendCommandList(songList, target, commandList);
            songList.clear();
        }
        ++Itr;
    }
    executeList(commandList);

    }
    else
    // dragging items around within the playlist
    if (srcView == playListView)
    {
        QListViewItem *thisItem;
        thisItem = srcView->firstChild();
    b_waitForShuffle = TRUE;
    LockStatUpdate();
    mpd_sendCommandListBegin(trConn);
    if (!ErrorCheck(153, "mpd_sendCommandListBegin"))
    {
        mpd_finishCommand(trConn);
        UnlockStatUpdate();
        b_waitForShuffle = FALSE;
        return;
    }
        while(thisItem != 0)
        {
            if(thisItem->isSelected())
        {
        thisItem->setSelected(FALSE);
        moveListItem(thisItem, target);
        thisItem = srcView->firstChild();
        continue;
        }
        else  // Move on to the next
            thisItem = thisItem->itemBelow();
        }
    mpd_sendCommandListEnd(trConn);
    mpd_finishCommand(trConn);
    ErrorCheck(153, "mpd_sendCommandListEnd");
    UnlockStatUpdate();
    b_waitForShuffle = FALSE;
    reNumberList();
    }
}


void qm_Tracks::moveListItem(QListViewItem *theItem, QListViewItem *target)
{
    int targetPos = 0;
    int startPos = theItem->itemPos() / theItem->height();
    crPlaylistIncrement();
    if(target)
    {
    targetPos = target->itemPos() / target->height();
    if (startPos == targetPos)
        return;
    QListViewItem *addAfter =  target->itemAbove();
    // inserting item in the list
    if (addAfter)
    {
        theItem->moveItem(addAfter);
            if(startPos < targetPos)
            targetPos--;
    }
    else
    // first item in the list
    {
        playListView->takeItem(theItem);
        playListView->insertItem(theItem);
    }
    }
    else
    {
        // last item in the list
    targetPos = playlistSize-1;
    QListViewItem *addAfter = playListView->lastItem();
    theItem->moveItem(addAfter);
    }


    mpd_sendMoveCommand(trConn, startPos, targetPos);
    mpd_finishCommand(trConn);
    ErrorCheck(154,"mpd_sendMoveCommand");
    // do not UnlockStatUpdate() here. This function is called from
    // within addDraggedItems(): the statuscheck must stay Locked!

    if (startPos == curSong)
    crCurSongChanged(curSong);
}


void qm_Tracks::addaDirectory(const QString & thePath, qm_SongPtrList & songList)
{
    if (trConn == NULL)
    return;
    mpd_InfoEntity *ntity;
    LockStatUpdate();
    // MPD iterates the item and returns ALL subitems.
    mpd_sendListallInfoCommand(trConn, thePath);
    if (ErrorCheck(155,"mpd_sendListallInfoCommand"))
    {
    while((ntity = mpd_getNextInfoEntity(trConn)) != NULL)
    {
        // Skip non-song items
        if(ntity->type != MPD_INFO_ENTITY_TYPE_SONG)
        {
            mpd_freeInfoEntity(ntity);
            continue;
        }
        mpd_Song *newSong = mpd_songDup (ntity->info.song);
        songList.append(newSong);
        mpd_freeInfoEntity(ntity);
    }
    }
    mpd_finishCommand(trConn);
    ErrorCheck(155,"mpd_sendListallInfoCommand");
    UnlockStatUpdate();
}


void qm_Tracks::addaPlaylist(const QString & listName)
{
    if (trConn == NULL)
    return;
    if (!b_CTRLpressed)
    clearPlayList();
    // Load the dropped list
    LockStatUpdate();
    mpd_sendLoadCommand(trConn, listName.utf8());
    mpd_finishCommand(trConn);
    ErrorCheck(156,"mpd_sendLoadCommand");
    UnlockStatUpdate();
    // The Core will notice the change and call trSetServerPlaylist
    // which will next set the track titles, playtime and number of tracks.
}


void qm_Tracks::addanArtist(const QString & artistName, qm_SongPtrList & songList)
{
    if (trConn == NULL)
    return;
    // get the album list
    QStringList albumList;
    char *albumName = NULL;
    QString theName;
    LockStatUpdate();

    // always check first for tracks without an album name
    albumList.append(" ??? ");

    mpd_sendListCommand (trConn, MPD_TABLE_ALBUM, artistName.utf8());
    if (ErrorCheck(103,"mpd_sendListCommand"))
    {
    while ((albumName = mpd_getNextAlbum(trConn)) != NULL)
    {
        theName = QString::fromUtf8(albumName);
        albumList.append(theName);
        free(albumName);
    }
    }
    mpd_finishCommand (trConn);
    ErrorCheck(103,"mpd_sendListCommand");
    UnlockStatUpdate();

    // add all found albums
    QStringList::Iterator AlbumIt;
    for (AlbumIt = albumList.begin(); AlbumIt != albumList.end(); ++AlbumIt )
    {
    addanAlbum(*AlbumIt, artistName, songList);
    }
}


void qm_Tracks::addanAlbum(const QString & albumName, const QString & theArtist, qm_SongPtrList & songList)
{
    if (trConn == NULL)
    return;
    mpd_InfoEntity *ntity = NULL;
    LockStatUpdate();

    qm_SongPtrList albumTracks;

   if (albumName.utf8() == " ??? ")
   {
    mpd_sendFindCommand (trConn, MPD_TABLE_ARTIST, theArtist.utf8() );
    if (ErrorCheck(105,"mpd_sendFindCommand"))
    {
        while ((ntity = mpd_getNextInfoEntity(trConn)) != NULL)
        {
            if(ntity->type == MPD_INFO_ENTITY_TYPE_SONG)
            {
                mpd_Song *newSong = mpd_songDup (ntity->info.song);
                // append if the song has no album name
                if ( (QString::fromUtf8(newSong->album)).isEmpty() )
                    albumTracks.append(newSong);
            }
            mpd_freeInfoEntity(ntity);
        }
    }
    mpd_finishCommand (trConn);
    ErrorCheck(106,"mpd_sendFindCommand");
   }
   else
   {
    mpd_sendFindCommand (trConn, MPD_TABLE_ALBUM, albumName.utf8());
    if (ErrorCheck(107,"mpd_sendFindCommand"))
    {
        while ((ntity = mpd_getNextInfoEntity(trConn)) != NULL)
        {
            if(ntity->type == MPD_INFO_ENTITY_TYPE_SONG)
            {
                mpd_Song *newSong = mpd_songDup (ntity->info.song);
                // Check if song belongs to the right artist! If not specified get them all
                if (theArtist == "not_specified" || theArtist == QString::fromUtf8(newSong->artist))
                    albumTracks.append(newSong);
            }
            mpd_freeInfoEntity(ntity);
        }
    }
    mpd_finishCommand (trConn);
    ErrorCheck(107,"mpd_sendFindCommand");
    }
    UnlockStatUpdate();

    // Bubblesort albumTracks by track number
    // Tracks without a track # come after the # tracks, in original oder (by filename).
    mpd_Song *curSong = 0;
    mpd_Song *prevSong = 0;
    int listsize = albumTracks.count();
    for (int i = (listsize - 1); i >= 0; i--)
    {
    bool nochange = TRUE;
    for (int j = 1; j <= i; j++)
    {
        curSong = albumTracks.at(j);
        int num_curr = ((QString)(curSong->track)).toInt(NULL, 10);
        if (!num_curr > 0)
            num_curr = 1000;
        prevSong = albumTracks.at(j-1);
        int num_prev = ((QString)(prevSong->track)).toInt(NULL, 10);
        if (!num_prev > 0)
            num_prev = 1000;

        if (num_prev > num_curr)
        {
            // take curSong out and insert it at prevSong's pos
            curSong = albumTracks.take(j);
            albumTracks.insert(j-1, curSong);
            nochange = FALSE;
        }
    }
    if (nochange)
        break;
     }

    // next append songs to songlist
    QPtrListIterator<mpd_Song> sit( albumTracks );
    mpd_Song *addSong;
    while ( (addSong = sit.current()) != 0 )
    {
    songList.append(addSong);
        ++sit;
    }
    albumTracks.clear();

}


void qm_Tracks::appendCommandList(qm_SongPtrList songList, QListViewItem *target,  qm_CommandList & commandList)
{
    int targetPos = 0;
    int inc = 0;
    mpd_Song *aSong;
    QListViewItem *addAfter = NULL;
    for ( aSong = songList.first(); aSong; aSong = songList.next() )
    {
        qm_ListItem *newItem;
    qm_CmdStruct newCommand;
    mpd_Song theSong = *aSong;
    int songTime = theSong.time;
    if (songTime < 0)
        songTime = 0;
    QString fileName = QString::fromUtf8(theSong.file);
        // Skip if, for some weird reason, there is no file
        if(fileName.isEmpty())
        continue;
    if(target)
    {
    // 'target' is 'pushed down' by newItems, which are consequently added in the correct order.
        addAfter = target->itemAbove();
        targetPos = target->itemPos() / target->height();
        if (addAfter)
            // add after the item above (previous newItem)
            newItem = new qm_ListItem(playListView, addAfter, theSong, trackIcon, "#", "", sTitle(theSong), formatTime(songTime) + " ");
        else
            // it's the first one in the list
            newItem = new qm_ListItem(playListView, theSong, trackIcon, "#", "", sTitle(theSong), formatTime(songTime) + " ");
    }
    else
        // Add it at the end
        newItem = new qm_ListItem(playListView, playListView->lastItem(), theSong, trackIcon, "#", "", sTitle(theSong), formatTime(songTime) + " ");

    newItem->setDragEnabled(TRUE);
    newItem->setType(LITEM_TRACK);
    if (target)
    {
        newCommand.cmd = QM_CMD_ADDMOV;
        newCommand.moveTo = targetPos++;
    }
    else
        newCommand.cmd = QM_CMD_ADD;
    newCommand.time = theSong.time;
    if (newCommand.time < 0)
        newCommand.time = 0;
    newCommand.file = QString::fromUtf8(theSong.file);
    commandList.append(newCommand);
    inc++;
    }
}

// Scrolls the list to show the currently playing song (else to the top)
void qm_Tracks::spaceKeyPressed()
{
    if(curSong != -1 && curSong < playlistSize)
    {
    QListViewItem *theSong;
    findPointer(curSong, playListView, theSong);
    playListView->ensureItemVisible(theSong);
    }
    else
    playListView->ensureItemVisible(playListView->firstChild());
}

// execute commandlist (created when adding, moving or deleting items).
void qm_Tracks::executeList(qm_CommandList  & theList)
{
    if(theList.isEmpty() || trConn == NULL)
    return;
    bool msgShown = FALSE;
    LockStatUpdate();
    mpd_sendCommandListBegin(trConn);
    if (!ErrorCheck(109,"mpd_sendCommandListBegin"))
    {
        mpd_finishCommand(trConn);
        theList.clear();
        UnlockStatUpdate();
    return;
    }
    qm_CommandList::iterator cit;
    for ( cit = theList.begin(); cit != theList.end(); ++cit )
    {
    qm_CmdStruct curCommand = *cit;
    switch (curCommand.cmd)
    {
        case QM_CMD_ADD:
        {
        //Check to see if we are trying to add too much
            if((plistMax > 0) && (playlistSize+1  > plistMax))
        {
            if (!msgShown)
            {
                QMessageBox::information( this,
                "Attention!", "The playlist has reached its max of " + QString::number(plistMax) + " items.\n"
                "One or more tracks were not added.\n");
                msgShown = TRUE;
            }
            continue;
        }
        ListTotalTime += curCommand.time;
        playlistSize++;
        crPlaylistIncrement();
        mpd_sendAddCommand(trConn, curCommand.file.utf8());
        mpd_finishCommand(trConn);
        ErrorCheck(158,"mpd_sendAddCommand");
        break;
       }
        case QM_CMD_ADDMOV:
        {
        //Check to see if we are trying to add too much
            if((plistMax > 0) && (playlistSize+1  > plistMax))
        {
            if (!msgShown)
            {
                QMessageBox::information( this,
                "Attention!", "The playlist has reached its max of " + QString::number(plistMax) + " items.\n"
                "One or more tracks were not added.\n");
                msgShown = TRUE;
            }
            continue;
        }
        ListTotalTime += curCommand.time;
        playlistSize++;
        crPlaylistIncrement();
        mpd_sendAddCommand(trConn, curCommand.file.utf8());
        mpd_finishCommand(trConn);
        ErrorCheck(159,"mpd_sendAddCommand");
        // move added track to right position
        crPlaylistIncrement();
        mpd_sendMoveCommand(trConn, playlistSize-1, curCommand.moveTo);
        mpd_finishCommand(trConn);
        ErrorCheck(159,"mpd_sendMoveCommand");
        break;
       }
       case QM_CMD_DEL:
       {
        ListTotalTime -= curCommand.time;
        crPlaylistIncrement();
        playlistSize--;
        mpd_sendDeleteCommand(trConn, curCommand.trackNr);
        mpd_finishCommand(trConn);
        ErrorCheck(160,"mpd_sendDeleteCommand");
        break;
       }
       default:
        continue;
    } // switch
    }
    mpd_sendCommandListEnd(trConn);
    mpd_finishCommand(trConn);

    if (!(ErrorCheck(109,"mpd_sendCommandListEnd")) )
    crUpdatePlaylist(); // reload playlist (and resetTime)
    else
    resetTime();

    UnlockStatUpdate();
    theList.clear();
    reNumberList();
}

// mpd_sendShuffleCommand is faster, but reloads list (without history tags)
void qm_Tracks::sendShuffle()
{
    if (playlistSize < 2)
    return;

    // do not allow trSetCurSong
    b_waitForShuffle = TRUE;
    bt_shuffle->setEnabled(FALSE);

    QListViewItem *randItem = NULL;
    QListViewItem *addAfter = NULL;
    int moveTo = 0;
    srand ( time(NULL) );
    LockStatUpdate();
    mpd_sendCommandListBegin(trConn);
    if (!ErrorCheck(165,"mpd_sendCommandListBegin"))
    {
    mpd_finishCommand(trConn);
        b_waitForShuffle = FALSE;
        bt_shuffle->setEnabled(TRUE);
    UnlockStatUpdate();
    return;
    }
    // move currently playing song (if so) to the top
    switch (curSong)
    {
    case -1: // No song playing. Get a random one and make it first.
    {
        int startPos = rand()%playlistSize;
        findPointer(startPos, playListView, randItem);
        if (startPos > 0 && randItem)
        {
            playListView->takeItem(randItem);
            playListView->insertItem(randItem);
            crPlaylistIncrement();
            mpd_sendMoveCommand(trConn, startPos, 0);
            mpd_finishCommand(trConn);
            ErrorCheck(162,"mpd_sendMoveCommand");
            break;
        }
    }
    case 0: // curSong is already first in the list.
    {
        break;
    }
    default: // curSong is down in the list. Get the pointer and make it first.
    {
        findPointer(curSong, playListView, randItem);
        if (randItem)
        {
            playListView->takeItem(randItem);
            playListView->insertItem(randItem);
            crPlaylistIncrement();
            mpd_sendMoveCommand(trConn, curSong, 0);
            mpd_finishCommand(trConn);
            ErrorCheck(163,"mpd_sendMoveCommand");
            curSong = 0;
        }
    }
    }
    // whatever now is firts in the list is addAfter.
    addAfter = playListView->firstChild();

    moveTo++;
    // randomly add the rest after addAfter
    while(moveTo < playlistSize)
    {
    int startPos = rand()%(playlistSize-moveTo) + moveTo;
    findPointer(startPos, playListView, randItem);
    if (startPos > moveTo && randItem)
    {
        randItem->moveItem(addAfter);
        crPlaylistIncrement(); // prevent reloading playlist
        mpd_sendMoveCommand(trConn, startPos, moveTo);
        mpd_finishCommand(trConn);
        ErrorCheck(164,"mpd_sendMoveCommand");
        addAfter = randItem;
    }
    // error or (startPos == moveTo), so move on
    else
        addAfter = addAfter->itemBelow();
    moveTo++;
    }
    mpd_sendCommandListEnd(trConn);
    mpd_finishCommand(trConn);
    ErrorCheck(165,"mpd_sendCommandListEnd");

    UnlockStatUpdate();
    reNumberList();
    bt_shuffle->setEnabled(TRUE);
    b_waitForShuffle = FALSE;
    // scroll to the top
    playListView->ensureItemVisible(playListView->firstChild());
}

// Clear the playlist (client and server side)
void qm_Tracks::clearPlayList()
{
    // client:
    playListView->clear();
    curSong = -1;
    playlistSize = 0;
    // server:
    LockStatUpdate();
    mpd_sendClearCommand(trConn);
    mpd_finishCommand(trConn);
    ErrorCheck(166,"mpd_sendClearCommand");
    UnlockStatUpdate();
}

// Get a song-title out of an mpd_Song
QString qm_Tracks::sTitle(mpd_Song theSong)
{
    QString sTitle = QString::fromUtf8(theSong.title);
    if(!sTitle.isEmpty())
    {
    QString sArtist = QString::fromUtf8(theSong.artist);
    if(!sArtist.isEmpty())
        return(sArtist + " - " + sTitle);
    else
        return(sTitle);
    }
    else
    {
    QString sFile = QString::fromUtf8(theSong.file);
    int i = sFile.findRev( '/' );
    return(sFile.right(sFile.length() - ++i));
    }
}

// Format time from a zillion secs to 000:00
QString qm_Tracks::formatTime(int playTime)
{
    if(playTime < 0 )
    playTime = 0;
    if(playTime == 0)
    return (QString("0:00"));
    int minutes = (int)(playTime / 60);
    int seconds = playTime % 60;
    QString time;
    if(seconds < 10)
    time = QString::number(minutes) + ":0" + QString::number(seconds);
    else
    time = QString::number(minutes) + ":" + QString::number(seconds);
    return time;
}

// Format time from a zillion secs to 000:00:00
QString qm_Tracks::listTime(int listTotalTime)
{
    if(listTotalTime < 0 )
    listTotalTime = 0;
    if(listTotalTime == 0)
    return (QString("0:00:00"));

    int hours = (int)(listTotalTime / 3600);
    int minutes = (int)((listTotalTime % 3600) / 60);
    int seconds = listTotalTime % 60;
    QString time;
    if(seconds < 10)
    time = QString::number(minutes) + ":0" + QString::number(seconds);
    else
    time = QString::number(minutes) + ":" + QString::number(seconds);
    if (hours != 0)
    {
    if (minutes < 10)
        time = QString::number(hours) + ":0" + time;
    else
        time = QString::number(hours) + ":" + time;
    }
    return time;
}

#endif
