#include <stdlib.h> // free
#include <qevent.h> // resize event
#include <qheader.h> // hide the column headers
#include <time.h> // rand (shuffle)
#include <qinputdialog.h> // saving playlist
#include <qfile.h> // reading mpd config file
#include <qmessagebox.h> // error msg etc
#include <qsettings.h> // to get the colorise option
#include <qmap.h> // to sort dragged albums before adding to list
#include <qstringlist.h> // the album list (iterated to add date).
#include <qapplication.h> // te get the app default font

// dataBaseView Modes
// (note that qm_Core.ui.h uses the INT values to init the tracks window)
#define DMODE_DIR 0
#define DMODE_ARTIST 1
#define DMODE_ALBUM 2
#define DMODE_LIST 3
#define DMODE_SEARCH 4
#define DMODE_GENRE 5

// extension of this file
#include "qm_Playlist.h"

bool b_trReallyClose;

void qm_Tracks::init()
{
    // caption icon
    setIcon(QPixmap::fromMimeSource( "mn_playlist.png" ));
    b_trReallyClose = FALSE;
    b_statusChecking = FALSE;
    b_noSearchNow = FALSE;
    b_commandBusy = FALSE;
    trConn = NULL;
    plistMax = 0;
    totalTracks = 0;
    // Setup the Playlist panel
    InitPlaylist();
    // types: 0 = accepts drops, 1 = allow drags, 2 = drag & drop
    dataBaseView->setType(LV_DRAG);

    // Setup left panel
    dataBaseView->addColumn(" Directory ", 0);
    dataBaseView->setAllColumnsShowFocus(TRUE);
    dataBaseView->setSelectionMode(QListView::Extended);
    dataBaseView->setHScrollBarMode(QScrollView::AlwaysOff);
    dataBaseView->setVScrollBarMode(QScrollView::AlwaysOn);
    dataBaseView->header()->setHidden(TRUE);

    artistIcon = QPixmap::fromMimeSource( "tr_artist.png" );
    trackIcon = QPixmap::fromMimeSource( "tr_track.png" );
    albumIcon = QPixmap::fromMimeSource( "tr_album.png" );
    folderIcon = QPixmap::fromMimeSource( "tr_folder.png" );
    crossIcon = QPixmap::fromMimeSource( "tr_redcross.png" );
    checkIcon = QPixmap::fromMimeSource( "tr_check.png" );
    playlistIcon = QPixmap::fromMimeSource( "tr_playlist.png" );
    trackPlayedIcon = QPixmap::fromMimeSource( "tr_trplayed.png" );
    plPlayIcon = QPixmap::fromMimeSource( "st_play.png" );
    plStopIcon = QPixmap::fromMimeSource( "st_stop.png" );
    plWaitIcon = QPixmap::fromMimeSource( "st_wait.png" );
    plPauseIcon = QPixmap::fromMimeSource( "st_pause.png" );

    // set colors
    QSettings settings;
    b_colorise = settings.readBoolEntry( "/quimup/General/colorise", TRUE );
    if (b_colorise)
    {
    QColor fCol = QColor(0, 10, 40);
    QPalette pal (bt_gosearch->palette()); // just grab any palette
    pal.setColor(QColorGroup::ButtonText, fCol);
    pal.setColor(QColorGroup::Text, fCol);
    pal.setColor(QColorGroup::Foreground, fCol);
    pal.setColor(QColorGroup::Background, QColor(218, 220, 224));
    pal.setColor(QColorGroup::Button, QColor(240, 245, 250));
    pal.setColor(QColorGroup::Highlight, QColor(75, 100, 125));
    pal.setColor(QColorGroup::Shadow, QColor(15, 30, 50));
    pal.setColor(QColorGroup::Mid, QColor(116, 130, 148));
    setPalette(pal); // set palette globally
    // Scrollbars don't get the 'button' color
    pal.setColor(QColorGroup::Button, QColor(218, 220, 224));
    dataBaseView->setPalette(pal);
    playListView->setPalette(pal);
        lb_tracks_time->setPaletteForegroundColor(QColor(65, 85, 105));
        fr_tracks_time->setPaletteForegroundColor(QColor(65, 85, 105));
        lb_dbstats->setPaletteForegroundColor(QColor(65, 85, 105));
    fr_dbstats->setPaletteForegroundColor(QColor(65, 85, 105));
   }

    QFont smallFont = QApplication::font();
    smallFont.setPointSize(smallFont.pointSize()-1);

    lb_tracks_time->setFont(smallFont);
    lb_dbstats->setFont(smallFont);
    bt_dbupd->setFont(smallFont);

    // Connection times out if idle, so send a signal once in a while
    keepAliveLoop = new QTimer(this);
    connect(keepAliveLoop, SIGNAL(timeout()), this, SLOT(StatusCheck()));
    connect(parent(), SIGNAL(trReloadData(bool)), this, SLOT(trReloadData(bool)));
    connect(parent(), SIGNAL(trSetServer( const QString &, int, const QString &)), this, SLOT(trSetServer( const QString &, int, const QString &)));
    connect(parent(), SIGNAL(trCloseConn()), this, SLOT(trCloseConn()));
    connect(parent(), SIGNAL(trReallyClose()), this, SLOT(trReallyClose()));
    connect(parent(), SIGNAL(trSetStatus(bool)), this, SLOT(trSetStatus(bool)));
    connect(parent(), SIGNAL(trSetData(int, bool)), this, SLOT(trSetData(int, bool)));
    connect(parent(), SIGNAL(trSetStatistics(int, int, int)), this, SLOT(trSetStatistics(int, int, int)));
    connect(parent(), SIGNAL(trSetListFonts(int)), this, SLOT(trSetListFonts(int)));
    connect(parent(), SIGNAL(trSetSplitter(int, int)), this, SLOT(trSetSplitter(int, int)));
}

void qm_Tracks::trSetListFonts(int listFsize)
{
    QSettings settings;
    QString listFont = settings.readEntry( "/quimup/General/listFont", "SansSerif" );
    QFont myFont;
    myFont.setFamily(listFont);
    myFont.setBold(FALSE);
    myFont.setPointSize(listFsize);
    dataBaseView->setFont(myFont);
    playListView->setFont(myFont);
}

void qm_Tracks::trSetData(int mode, bool his)
{
    dataMode = mode;
    cb_mode->setCurrentItem(mode);
    bx_markplayed->setChecked(his);
    b_markplayed = his;
}


void qm_Tracks::reloadClicked()
{
    trReloadData(TRUE);
}


void qm_Tracks::trReloadData(bool getmode)
{
    // TRUE when called from mode combobox
    if (getmode)
    {
    dataMode = cb_mode->currentItem();
    // Store in the Core
    crGetsTrData(dataMode, b_markplayed);
        // reset selection mode in case we come from the genre list
        dataBaseView->setSelectionMode(QListView::Extended);
    }

    dataBaseView->clear();

    switch (dataMode)
    {
        case DMODE_DIR:
        expandDirectory(NULL);
        break;
        case DMODE_ARTIST:
        expandArtist(NULL);
        break;
        case DMODE_ALBUM:
        expandAlbum(NULL);
        break;
        case DMODE_LIST:
        expandPlaylist(NULL);
        break;
    case DMODE_SEARCH:
        search();
        break;
    default:
        return;
    }
}


void qm_Tracks::expandListItem(QListViewItem * theItem)
{
    switch (dataMode)
    {
        case DMODE_DIR:
        expandDirectory((qm_ListItem*)theItem);
        break;
        case DMODE_ARTIST:
        expandArtist((qm_ListItem*)theItem);
        break;
        case DMODE_ALBUM:
        expandAlbum((qm_ListItem*)theItem);
        break;
        case DMODE_LIST:
        expandPlaylist((qm_ListItem*)theItem);
        break;
    default:
        break;
    }
}


void qm_Tracks::expandPlaylist(qm_ListItem * theItem)
{
    if(trConn==NULL)
    return;
    // vars
    int type;
    // getting the RoOt items
    if (theItem == NULL)
    {
    type = LITEM_ROOT;
    theItem = new qm_ListItem(dataBaseView, PlayListPath,  LITEM_STAT_HEADER);
    theItem->setDragEnabled(FALSE);
    theItem->setSelectable(FALSE);
    theItem->setPixmap(0,checkIcon);
    theItem->setExpandable(TRUE);
    }
    else
    {
    // sub-items already added: nothing to do
    if (theItem->firstChild())
        return;
    type = theItem->getType();
    }
    // all set to go
    LockStatUpdate();

    switch (type)
    {
    case LITEM_ROOT:
    {
      // get the titles
      qm_ListItem * ItemToAdd;
      mpd_InfoEntity *ntity = NULL;
      int count = 0;
      mpd_sendLsInfoCommand(trConn, "");
      bool success = ErrorCheck(2,"mpd_sendLsInfoCommand");
      if (success)
      {
        while ((ntity = mpd_getNextInfoEntity(trConn)) != NULL)
        {
            if(ntity->type == MPD_INFO_ENTITY_TYPE_PLAYLISTFILE)
            {
                mpd_PlaylistFile *newList = mpd_playlistFileDup (ntity->info.playlistFile);
                // get the list's name
                QString listName = QString::fromUtf8(newList->path);
                if (listName.isEmpty())
                    listName = "?";
                ItemToAdd = new qm_ListItem(theItem, *newList, listName);
                ItemToAdd->setAlbum(listName); // use 'album' to store the list-name
                ItemToAdd->setType(LITEM_PLAYLIST);
                ItemToAdd->setDragEnabled(TRUE);
                ItemToAdd->setExpandable(TRUE);
                ItemToAdd->setPixmap(0,playlistIcon);
                count++;
            }
            mpd_freeInfoEntity(ntity);
          }
      }
      mpd_finishCommand (trConn);
      if (success)
        success = ErrorCheck(2,"mpd_sendLsInfoCommand");

      if (count == 0 || !success)
      {
        QListViewItem *itemPointer = dataBaseView->firstChild();
        if (!success)
            itemPointer->setText(0,"Error getting playlists");
        else
            itemPointer->setText(0,"0 playlists");
        itemPointer->setSelectable(FALSE);
        itemPointer->setSelected(FALSE);
        itemPointer->setDragEnabled(FALSE);
        itemPointer->setPixmap(0,crossIcon);
      }
      else
        theItem->setOpen(TRUE);
      break;
    }

    case LITEM_PLAYLIST:
    {
        qm_ListItem * ItemToAdd;
        if (PlayListPath.isEmpty())
        {
            ItemToAdd = new qm_ListItem(theItem, "Could not get path", LITEM_STAT_MESSAGE);
            ItemToAdd->setSelectable(FALSE);
            break;
        }

        QFile file(PlayListPath + theItem->text(0) + ".m3u");
        int count = 0;
        if (file.open( IO_ReadOnly ) )
        {
            QTextStream stream( &file );
            while ( !stream.atEnd() && count < 25)
            {
                count++;
                QString line = QString::number(count);
                if (count < 10)
                    line = "0" + line;
                line += "  ";
                if (count < 25)
                    line += nameOf(stream.readLine());
                else
                    line = " ...etc.";
                // append after lastitem
                ItemToAdd = new qm_ListItem(theItem, dataBaseView->lastItem(), line, LITEM_STAT_PLITEM);
                ItemToAdd->setPixmap(0,trackIcon);
                ItemToAdd->setSelectable(FALSE);
            }
            file.close();
        }
        else
        {
            ItemToAdd = new qm_ListItem(theItem, "Could not open file", LITEM_STAT_MESSAGE);
            ItemToAdd->setSelectable(FALSE);
        }
      break;
    }

    default:
    {
      // nothing to do
      UnlockStatUpdate();
      return;
    }
    }
   UnlockStatUpdate();
}


// Create or expand artist items
void qm_Tracks::expandArtist(qm_ListItem * theItem)
{
    if(trConn == NULL)
    return;
    // vars
    int type;
    // getting the RoOt items
    if (theItem == NULL)
    {
    type = LITEM_ROOT;
    theItem = new qm_ListItem(dataBaseView, "Artists",  LITEM_STAT_HEADER);
    theItem->setDragEnabled(FALSE);
    theItem->setSelectable(FALSE);
    theItem->setPixmap(0,checkIcon);
    theItem->setExpandable(TRUE);
    }
    else
    // getting sub-items
    {
    // sub-items already added: nothing to do
    if (theItem->firstChild() )
        return;
    type = theItem->getType();
    }

    // all set to go
    LockStatUpdate();

    switch (type)
    {
    case LITEM_ROOT:
    {
      // get the artists list
      char *artistName = NULL;
      QString theName;
      qm_ListItem * ItemToAdd;
      int count = 0;
      mpd_sendListCommand (trConn, MPD_TABLE_ARTIST, NULL);
      bool success = ErrorCheck(4,"mpd_sendLsInfoCommand");
      if (success)
      {
          while ((artistName = mpd_getNextArtist(trConn)) != NULL)
          {
            theName = QString::fromUtf8(artistName);
            if (theName.isEmpty() || theName == " ")
                theName = "?";
            ItemToAdd = new qm_ListItem(theItem, theName, LITEM_STAT_NORMAL);
            ItemToAdd->setArtist(theName);
            ItemToAdd->setType(LITEM_ARTIST);
            ItemToAdd->setExpandable(TRUE);
            ItemToAdd->setPixmap(0,artistIcon);
            ItemToAdd->setDragEnabled(TRUE);
            free(artistName);
            count++;
          } // end while
      }
      mpd_finishCommand (trConn);
      if (success)
        success = ErrorCheck(4,"mpd_sendLsInfoCommand");

      if (count == 0 || !success)
      {
        QListViewItem *itemPointer = dataBaseView->firstChild();
        if (!success)
            itemPointer->setText(0,"Error getting artists");
        else
            itemPointer->setText(0,"0 artists");
        itemPointer->setPixmap(0,crossIcon);
      }
      else
      {
        QListViewItem *itemPointer = dataBaseView->firstChild();
        itemPointer->setText(0, QString::number(count) + " artists");
        theItem->setOpen(TRUE);
      }

      break;
    }

    case LITEM_ARTIST:
    {
        qm_ListItem * ItemToAdd;
        mpd_InfoEntity *ntity = NULL;
        QString currentArtist = theItem->getArtist();
        int count = 0;
            // Firts check if there is a track (or more) for this artist that has no album name.
            // If so, the album " ??? " will be added.
        mpd_sendFindCommand (trConn, MPD_TABLE_ARTIST, currentArtist.utf8() );
        if (ErrorCheck(6,"mpd_getNextInfoEntity"))
        {
            while ((ntity = mpd_getNextInfoEntity(trConn)) != NULL)
            {
                if(ntity->type == MPD_INFO_ENTITY_TYPE_SONG)
                {
                    mpd_Song *newSong = mpd_songDup (ntity->info.song);
                    if ((QString::fromUtf8(newSong->album)).isEmpty())
                    {
                        ItemToAdd = new qm_ListItem(theItem, " ??? ", LITEM_STAT_NORMAL);
                        ItemToAdd->setAlbum(" ??? ");
                        ItemToAdd->setType(LITEM_ALBUM);
                        ItemToAdd->setArtist(currentArtist);
                        ItemToAdd->setExpandable(TRUE);
                        ItemToAdd->setPixmap(0,albumIcon);
                        ItemToAdd->setDragEnabled(TRUE);
                        mpd_freeInfoEntity(ntity);
                        count++;
                        break;
                    }
                }
            mpd_freeInfoEntity(ntity);
            }
        }
        mpd_finishCommand (trConn);
        ErrorCheck(6,"mpd_getNextInfoEntity");


        // Now get the albums for this artist and put the names in a list
        QStringList albumList;
        char *albumName = NULL;
        mpd_sendListCommand (trConn, MPD_TABLE_ALBUM, currentArtist.utf8() );
        if (ErrorCheck(7,"mpd_sendListCommand"))
        {
            while ((albumName = mpd_getNextAlbum(trConn)) != NULL)
            {
                QString newAlbum = QString::fromUtf8(albumName);
                    if (!newAlbum.isEmpty())
                    albumList.append(QString::fromUtf8(newAlbum));
                free(albumName);
                count++;
            }
        }
        mpd_finishCommand (trConn);
        ErrorCheck(7,"mpd_sendListCommand");


          // Next iterate the list and get the date for each album
          for ( QStringList::Iterator itr = albumList.begin(); itr != albumList.end(); ++itr )
          {
            QString theAlbum = (*itr).utf8();
            QString date = ".... ";

            // try to find a date in the tracks
            mpd_InfoEntity *ntity = NULL;
            mpd_sendFindCommand (trConn, MPD_TABLE_ALBUM, theAlbum.utf8());
            if (ErrorCheck(8,"mpd_sendFindCommand"))
            {
                while ((ntity = mpd_getNextInfoEntity(trConn)) != NULL)
                {
                    if(ntity->type == MPD_INFO_ENTITY_TYPE_SONG)
                    {
                        mpd_Song *newSong = mpd_songDup (ntity->info.song);
                        // different artists can have albums with the same title!!
                        if ( currentArtist != QString::fromUtf8(newSong->artist))
                        {
                            mpd_freeInfoEntity(ntity);
                            continue;
                        }
                        QString date_found = QString::fromUtf8(newSong->date);
                        if (!date_found.isEmpty())
                        {
                            date = date_found + "  ";
                            mpd_freeInfoEntity(ntity);
                            break;
                        }
                    }
                    mpd_freeInfoEntity(ntity);
                }
            }
            mpd_finishCommand (trConn);
            ErrorCheck(8,"mpd_sendFindCommand");

            QString Date_Album = date + theAlbum;
            ItemToAdd = new qm_ListItem(theItem, Date_Album, LITEM_STAT_NORMAL);
            ItemToAdd->setAlbum(theAlbum);
            ItemToAdd->setType(LITEM_ALBUM);
            ItemToAdd->setArtist(currentArtist);
            ItemToAdd->setExpandable(TRUE);
            ItemToAdd->setPixmap(0,albumIcon);
            ItemToAdd->setDragEnabled(TRUE);
          }

          theItem->setOpen(TRUE);
          break;
    }

    case LITEM_ALBUM:
    {
       // get the titles
      qm_ListItem * ItemToAdd;
      mpd_InfoEntity *ntity = NULL;
      QString currentArtist = theItem->getArtist();
      QString alBum = theItem->getAlbum();
      // get the tracks that have no album name

      if (alBum == " ??? ")
      {
        mpd_sendFindCommand (trConn, MPD_TABLE_ARTIST, currentArtist.utf8() );
        if (ErrorCheck(11,"mpd_sendFindCommand"))
        {
            while ((ntity = mpd_getNextInfoEntity(trConn)) != NULL)
            {
                if(ntity->type == MPD_INFO_ENTITY_TYPE_SONG)
                {
                    mpd_Song *newSong = mpd_songDup (ntity->info.song);
                    // do nothing if album-name is present
                    if (!(QString::fromUtf8(newSong->album)).isEmpty())
                    {
                        mpd_freeInfoEntity(ntity);
                        continue;
                    }

                    // get the song's name
                    QString songName = QString::fromUtf8(ntity->info.song->title);
                    if (songName.isEmpty())
                        songName = "?";
                    else
                    {
                        QString trNum = QString::fromUtf8(ntity->info.song->track);
                        if (!trNum.isEmpty())
                        {
                            // remove the x/xx format
                            int i = trNum.findRev( '/' );
                            if (i != -1)
                                trNum = trNum.left(i);
                            // add leading zero, when required
                            if (trNum.length() == 1)
                                trNum = "0" + trNum;
                        }
                        else
                        trNum = "?";
                        songName = trNum + "  " + songName;
                    }
                    ItemToAdd = new qm_ListItem(theItem, *newSong, songName);
                    ItemToAdd->setType(LITEM_TRACK);
                    ItemToAdd->setDragEnabled(TRUE);
                    ItemToAdd->setExpandable(FALSE);
                    ItemToAdd->setPixmap(0,trackIcon);
                }
                mpd_freeInfoEntity(ntity);
            }
        }
        mpd_finishCommand (trConn);
        ErrorCheck(11,"mpd_sendFindCommand");
        break;
          }

    mpd_sendFindCommand (trConn, MPD_TABLE_ALBUM, alBum.utf8() );
    if (ErrorCheck(12, "mpd_sendFindCommand"))
    {
        while ((ntity = mpd_getNextInfoEntity(trConn)) != NULL)
        {
            if(ntity->type == MPD_INFO_ENTITY_TYPE_SONG)
            {
                mpd_Song *newSong = mpd_songDup (ntity->info.song);
                // different artists can have albums with the same title!!
                if (currentArtist != QString::fromUtf8(newSong->artist))
                {
                    mpd_freeInfoEntity(ntity);
                    continue;
                }
                // get the song's name
                QString songName = QString::fromUtf8(ntity->info.song->title);
                if (songName.isEmpty())
                    songName = "?"; //nameOf(QString::fromUtf8(ntity->info.song->file));
                else
                {
                    QString trNum = QString::fromUtf8(ntity->info.song->track);
                    if (!trNum.isEmpty())
                    {
                        // remove the x/xx format
                        int i = trNum.findRev( '/' );
                        if (i != -1)
                            trNum = trNum.left(i);
                        // add leading zero, when required
                        if (trNum.length() == 1)
                            trNum = "0" + trNum;
                    }
                    else
                    trNum = "?";
                    songName = trNum + "  " + songName;
                }
                ItemToAdd = new qm_ListItem(theItem, *newSong, songName);
                ItemToAdd->setType(LITEM_TRACK);
                ItemToAdd->setDragEnabled(TRUE);
                ItemToAdd->setExpandable(FALSE);
                ItemToAdd->setPixmap(0,trackIcon);
            }
            mpd_freeInfoEntity(ntity);
          }
    }
    mpd_finishCommand (trConn);
    ErrorCheck(13, "mpd_sendFindCommand");
    break;
    }

    default:
    {
      // nothing to do
      UnlockStatUpdate();
      return;
    }
  } // end switch
  UnlockStatUpdate();
}


// Create or expand album items
void qm_Tracks::expandAlbum(qm_ListItem * theItem)
{
    if(trConn == NULL)
    return;
    // vars
    int type;
    // getting the RoOt items
    if (theItem == NULL)
    {
    type = LITEM_ROOT;
    theItem = new qm_ListItem(dataBaseView, "Albums",  LITEM_STAT_HEADER);
    theItem->setDragEnabled(FALSE);
    theItem->setSelectable(FALSE);
    theItem->setPixmap(0,checkIcon);
    theItem->setExpandable(TRUE);
    }
    else
    // getting sub-items
    {
    // sub-items already added: nothing to do
    if (theItem->firstChild() )
        return;
    type = theItem->getType();
    }

    // all set to go
    LockStatUpdate();

    switch (type)
    {
    case LITEM_ROOT:
    {
        qm_ListItem * ItemToAdd;
        QString theAlbum;
        int count = 0;
        char *albumName = NULL;
        mpd_sendListCommand (trConn, MPD_TABLE_ALBUM, NULL);
        bool success = ErrorCheck(16, "mpd_sendListCommand");
        if (success)
        {
            while ((albumName = mpd_getNextAlbum(trConn)) != NULL)
            {
                theAlbum = QString::fromUtf8(albumName);
                ItemToAdd = new qm_ListItem(theItem, theAlbum, LITEM_STAT_NORMAL);
                ItemToAdd->setAlbum(theAlbum);
                ItemToAdd->setArtist("not_specified");
                ItemToAdd->setType(LITEM_ALBUM);
                ItemToAdd->setExpandable(TRUE);
                ItemToAdd->setPixmap(0,albumIcon);
                ItemToAdd->setDragEnabled(TRUE);
                free(albumName);
                count++;
            } // end while
        }
        mpd_finishCommand (trConn);
        if (success)
            success = ErrorCheck(16, "mpd_sendListCommand");

        if (count == 0 || !success)
        {
            QListViewItem *itemPointer = dataBaseView->firstChild();
            if (!success)
                itemPointer->setText(0,"Error getting albums");
            else
                itemPointer->setText(0,"0 albums");
            itemPointer->setPixmap(0,crossIcon);
        }
        else
        {
            QListViewItem *itemPointer = dataBaseView->firstChild();
            itemPointer->setText(0, QString::number(count) + " albums");
            theItem->setOpen(TRUE);
        }
      break;
    }


    case LITEM_ALBUM:
    {
       // get the titles
      qm_ListItem * ItemToAdd;
      mpd_InfoEntity *ntity = NULL;
      mpd_sendFindCommand (trConn, MPD_TABLE_ALBUM, (theItem->getAlbum()).utf8() );
      if (ErrorCheck(17,"mpd_sendFindCommand"))
      {
        while ((ntity = mpd_getNextInfoEntity(trConn)) != NULL)
        {
            if(ntity->type == MPD_INFO_ENTITY_TYPE_SONG)
            {
                mpd_Song *newSong = mpd_songDup (ntity->info.song);
                // different artists can have albums with the same title.
                // but we are not in 'artist' mode, and below we add the artist's name
                // to the track title, so the difference will be obvious

                // get the song's name
                QString songName = QString::fromUtf8(ntity->info.song->title);
                if (songName.isEmpty())
                    songName = "?"; //nameOf(QString::fromUtf8(ntity->info.song->file));
                else
                {
                    QString trNum = QString::fromUtf8(ntity->info.song->track);
                    if (!trNum.isEmpty())
                    {
                        // remove the x/xx format
                        int i = trNum.findRev( '/' );
                        if (i != -1)
                            trNum = trNum.left(i);
                        // add leading zero, when required
                        if (trNum.length() == 1)
                            trNum = "0" + trNum;
                    }
                    else
                    trNum = "?";

                    QString arTist = QString::fromUtf8(newSong->artist);
                    if (arTist.isEmpty())
                        arTist = " ??? ";
                    songName = arTist + " - " + trNum + "  " + songName;
                }
                ItemToAdd = new qm_ListItem(theItem, *newSong, songName);
                ItemToAdd->setType(LITEM_TRACK);
                ItemToAdd->setDragEnabled(TRUE);
                ItemToAdd->setExpandable(FALSE);
                ItemToAdd->setPixmap(0,trackIcon);
            }
            mpd_freeInfoEntity(ntity);
          }
    }
    mpd_finishCommand (trConn);
    ErrorCheck(17,"mpd_sendFindCommand");
    break;
    }

    default:
    {
      // nothing to do
      UnlockStatUpdate();
      return;
    }
    } // end switch
    UnlockStatUpdate();
}




// Create or expand directory items
void qm_Tracks::expandDirectory(qm_ListItem * theItem)
{
    if(trConn == NULL)
    return;

    // Don't even try this with an empty database.
    // MPD can handle it, but quimup segfaults. Here's a nifty work-around:
    if (totalTracks == 0)
    {
    theItem = new qm_ListItem(dataBaseView, "0 items",  LITEM_STAT_HEADER);
    theItem->setDragEnabled(FALSE);
    theItem->setExpandable(FALSE);
    theItem->setSelectable(FALSE);
    theItem->setType(-1);
    theItem->setPixmap(0,crossIcon);
    return;
    }

    // vars
    qm_ListItem * ItemToAdd;
    mpd_InfoEntity * ntity;
    QString dirPath;
    bool makeRoOt = FALSE;

    // getting the RoOt items
    if (theItem == NULL)
    {
    makeRoOt = TRUE;

    theItem = new qm_ListItem(dataBaseView, MusicPath,  LITEM_STAT_HEADER);
    theItem->setDragEnabled(FALSE);
    theItem->setSelectable(FALSE);
    theItem->setPixmap(0,checkIcon);
    theItem->setExpandable(TRUE);

    dirPath = "";
    }
    // expanding a directory
    else
    {
    // sub-items already added or it's a track
    if (theItem->firstChild() || theItem->getType() != LITEM_DIR)
        return;
    // get the path to this dir
    mpd_Directory theDir = theItem->getDir();
    dirPath = theDir.path;
        // This comes straight from mpd and is utf8, which is what we need
        // below for mpd_sendLsInfoCommand. No need to convert.
    }

    // all set to go
    LockStatUpdate();

    mpd_sendLsInfoCommand(trConn, dirPath);
    if (ErrorCheck(19, "pd_sendLsInfoCommand"))
    {
    while((ntity = mpd_getNextInfoEntity(trConn)) != NULL)
    {
        if(ntity->type == MPD_INFO_ENTITY_TYPE_DIRECTORY)
        {
            mpd_Directory *newDir = mpd_directoryDup (ntity->info.directory);
            ItemToAdd = new qm_ListItem(theItem, *newDir, nameOf(QString::fromUtf8(ntity->info.directory->path)));
            ItemToAdd->setDragEnabled(TRUE);
            // any dir in the database should have songs in it:
            ItemToAdd->setExpandable(TRUE);
            ItemToAdd->setType(LITEM_DIR);
            ItemToAdd->setPixmap(0,folderIcon);
        }
        else
        if(ntity->type == MPD_INFO_ENTITY_TYPE_SONG)
        {
            mpd_Song *newSong = mpd_songDup (ntity->info.song);
            ItemToAdd = new qm_ListItem(theItem, *newSong, nameOf(QString::fromUtf8(ntity->info.song->file)));
            ItemToAdd->setType(LITEM_TRACK);
            ItemToAdd->setDragEnabled(TRUE);
            ItemToAdd->setExpandable(FALSE);
            ItemToAdd->setPixmap(0,trackIcon);
        }
        else
        {
            // No file, no dir. That's weird. Skip!
            mpd_freeInfoEntity(ntity);
            continue;
        }
        mpd_freeInfoEntity(ntity);
    } // end while
    }
    mpd_finishCommand (trConn);
    ErrorCheck(19, "pd_sendLsInfoCommand");

    UnlockStatUpdate();
   // open the RoOt
    if (makeRoOt)
    theItem->setOpen(TRUE);
}


// Unselect items in a dir that collapses or is selected
void qm_Tracks::unSelectSubs(QListViewItem * theItem)
{
    QListViewItem *subItem = theItem->firstChild();
    while (subItem)
    {
      if(subItem->isSelected())
    {
        subItem->setSelected(FALSE);
        dataBaseView->repaintItem(subItem);
    }
        if (subItem->firstChild())
    {
        unSelectSubs(subItem); // Auto-iterate sub-items
    }
    subItem = subItem->nextSibling();
    }
}


// Signal from the Core: give trWindow its own connection
void qm_Tracks::trSetServer( const QString & srvName, int srvPort , const QString & srvPassword)
{
    LockStatUpdate();
    // Close existing
    if(trConn != NULL)
    {
    mpd_clearError(trConn);
    mpd_closeConnection(trConn);
    trConn = NULL;
    }

    trConn = mpd_newConnection (srvName, srvPort, 5);

    // If connection failed we try once more after 0.25 sec
    if (trConn->error || trConn == NULL)
    {
    if (trConn != NULL)
    {
            mpd_clearError(trConn);
        mpd_closeConnection(trConn);
        trConn = NULL;
    }
    // wait 250 msec
        usleep(250000);
    trConn = mpd_newConnection (srvName, srvPort, 5);
    }

    if(!ErrorCheck(21,"mpd_newConnection"))
    {
    UnlockStatUpdate();
    return;
    }

    if(!srvPassword.isEmpty())
    {
    mpd_sendPasswordCommand(trConn, srvPassword);
    mpd_finishCommand(trConn);
    if (!ErrorCheck(22,"mpd_sendPasswordCommand"))
        {
        UnlockStatUpdate();
        return;
        }
    }
    // else we are connected
    UnlockStatUpdate();
    // Reset the tracks list (make sure dataMode is set first)
    trReloadData(FALSE);
    curSong = -1;

    b_waitForList = TRUE; // see qm_Playlist.h
}


bool qm_Tracks::ErrorCheck(int nr, QString msg)
{
    // no connection
    if(trConn == NULL)
    {
    // Signal the Core
    crConnectError();
    printf("Tracks Error: No Connection.\n");
    return FALSE;
    }

    // If an error occurred
    if(trConn->error)
    {
    // If it is a non-fatal error (no permission etc.).
    if(trConn->error == MPD_ERROR_ACK)
    {
        QString sError = QString::fromUtf8(trConn->errorStr);
            printf("Tracks Ack Error: " + QString::number(nr) + " " + msg + "\n" + sError + "\n");
        mpd_clearError(trConn);
        return TRUE;
    }

    // The error is fatal: disconnnect.
    mpd_clearError(trConn);
    mpd_closeConnection(trConn);
    trConn =NULL;
    // datBase time out
    keepAliveLoop->stop();
    // Signal the Core
    crConnectError();
    printf("Tracks Fatal Error " + QString::number(nr) + " " + msg + "\n");
    return FALSE;
    }

    // Nothinh wrong, apparently
    mpd_clearError(trConn);
    return TRUE;
}


// Strips the path off a filename
QString qm_Tracks::nameOf(const QString & thePath)
{
    // find the last /
    int i = thePath.findRev( '/' );
    // return everything to the right of it
    return(thePath.right(thePath.length() - ++i));
}


// called to keep the connection from timing out
void qm_Tracks::StatusCheck()
{
    if (!ErrorCheck(23,"Starting Tracks StatusCheck"))
       return;
    // Skip cycle if - previous has not finished - a command is busy
    if(b_commandBusy || b_statusChecking)
    {
    printf("Tracks skipped a statuscheck\n");
    return;
    }
    b_statusChecking = TRUE;
    // Let MPD know we're still here
    mpd_sendStatusCommand(trConn);
    mpd_finishCommand(trConn);
    // Unless there is an error
    if(!ErrorCheck(24,"Close Tr StatusCheck"))
    {
    trCloseConn();
    // and ask the Core for help
    crConnectError();
    }
    // Continue
    b_statusChecking = FALSE;
}

// Called by the Core when connection is lost
void qm_Tracks::trCloseConn()
{
    keepAliveLoop->stop();
    if(trConn) mpd_closeConnection(trConn);
    trConn =NULL;
}


// Set the the widgets
void qm_Tracks::trSetStatus(bool status)
{
    //We just set all of the widgets enabledness to that of status
    dataBaseView->setEnabled(status);
    cb_mode->setEnabled(status);
    cb_searchtype->setEnabled(status);
    bt_gosearch->setEnabled(status);
    le_searchstring->setEnabled(status);
    bt_dbupd->setEnabled(status);
    // Playlist:
    playListView->setEnabled(status);
    bt_shuffle->setEnabled(status);
    bt_clear->setEnabled(status);
    bt_purge->setEnabled(status);
    bt_save->setEnabled(status);
    lb_tracks_time->setEnabled(status);
}


// Called when this window is closed
void qm_Tracks::closeEvent( QCloseEvent * a )
{
    if(!b_trReallyClose)
    {
    // don't close, just toggle hide
    crToggleTracks();
    }
    else
    {
    // save the splitter settings;
    QSettings settings;
        settings.writeEntry( "/quimup/General/dbasePanelW", dbasePanelWidth );
        settings.writeEntry( "/quimup/General/plistPanelW", plistPanelWidth );
    }
    // and accept close
    a->accept();

}

// Signal from the Core.
void qm_Tracks::trReallyClose()
{
    // This time allow closing,
    b_trReallyClose = TRUE;
    // and trigger closeEvent.
    this->close();
}

void qm_Tracks::LockStatUpdate()
{
    while (b_commandBusy)
    {
    // Wait for previous command to finish
    }

    while(b_statusChecking)
    {
    // wait for statuscheck to finish
    }

    // and (b)lock the keepAliveLoop
    b_commandBusy = TRUE;
}


// Connection times out after 5 seconds, so:
void qm_Tracks::UnlockStatUpdate()
{
    // un(b)lock the statusloop
    b_commandBusy = FALSE;
    b_statusChecking = FALSE;
    if (!(keepAliveLoop->isActive ()) && trConn != NULL)
    keepAliveLoop->start(2000);
}


void qm_Tracks::search()
{
    if(trConn == NULL || b_noSearchNow)
    return;

    // Text going into mpd is converted to utf8. Text coming out is converted from utf8.
    // The searchstring does not go into mpd, but is matched with e.g. albums that come out.
    // Except for 'title' and 'genre'. So we apply .utf8() to the searchString only there.
    QString  searchString = le_searchstring->text();
    if(searchString.isEmpty())
    {
    dataBaseView->clear();
    qm_ListItem *RoOt = new qm_ListItem(dataBaseView, "Nothing to search for!",  LITEM_STAT_HEADER);
    RoOt->setDragEnabled(FALSE);
    RoOt->setSelectable(FALSE);
    RoOt->setPixmap(0,crossIcon);
    RoOt->setExpandable(FALSE);
        dataMode = DMODE_SEARCH;
    cb_mode->setCurrentItem(DMODE_SEARCH);
    return;
    }

    b_noSearchNow = TRUE;

    // reset selection mode in case we come from the genre list
    dataBaseView->setSelectionMode(QListView::Extended);
    // set database mode
    cb_mode->setCurrentItem(DMODE_SEARCH);

    dataBaseView->clear();
    qm_ListItem *RoOt = new qm_ListItem(dataBaseView, "Search",  LITEM_STAT_HEADER);
    RoOt->setPixmap(0,checkIcon);
    RoOt->setSelectable(FALSE);
    RoOt->setDragEnabled(FALSE);
    RoOt->setExpandable(TRUE);

    int searchMode = cb_searchtype->currentItem();
    int count = 0;


    if (searchMode == 0) //// Artist ////
    {
        LockStatUpdate();
    dataMode = DMODE_ARTIST;
    // get the artists list
    QString theArtist;
    qm_ListItem * ItemToAdd;
    char *artistName = NULL;
    mpd_sendListCommand (trConn, MPD_TABLE_ARTIST, NULL);
    if (ErrorCheck(26,"mpd_sendListCommand"))
    {
        while ((artistName = mpd_getNextArtist(trConn)) != NULL)
        {
            theArtist = QString::fromUtf8(artistName);
            if (theArtist.contains(searchString, FALSE))
            {
                ItemToAdd = new qm_ListItem(RoOt, theArtist, LITEM_STAT_NORMAL);
                ItemToAdd->setArtist(theArtist);
                ItemToAdd->setType(LITEM_ARTIST);
                ItemToAdd->setExpandable(TRUE);
                ItemToAdd->setPixmap(0,artistIcon);
                ItemToAdd->setDragEnabled(TRUE);
                count++;
            }
            free(artistName);
        } // end while
    }
    mpd_finishCommand (trConn);
    ErrorCheck(26,"mpd_sendListCommand");

    if (count)
    {
        QListViewItem *itemPointer = dataBaseView->firstChild();
        itemPointer->setText(0,QString::number(count) + " artists");
    }
      UnlockStatUpdate();
    }

    else
    if (searchMode == 1) //// Album ////
    {
        dataMode = DMODE_ALBUM;
        LockStatUpdate();
        qm_ListItem * ItemToAdd;
        QString theAlbum;
        char *albumName = NULL;
        mpd_sendListCommand (trConn, MPD_TABLE_ALBUM, NULL);
        if (ErrorCheck(28,"mpd_sendListCommand"))
        {
            while ((albumName = mpd_getNextAlbum(trConn)) != NULL)
            {
                theAlbum = QString::fromUtf8(albumName);
                if (theAlbum.contains(searchString, FALSE))
                {
                    ItemToAdd = new qm_ListItem(RoOt, theAlbum, LITEM_STAT_NORMAL);
                    ItemToAdd->setAlbum(theAlbum);
                    ItemToAdd->setArtist("not_specified");
                    ItemToAdd->setType(LITEM_ALBUM);
                    ItemToAdd->setExpandable(TRUE);
                    ItemToAdd->setPixmap(0,albumIcon);
                    ItemToAdd->setDragEnabled(TRUE);
                    count++;
                }
                free(albumName);
            } // end while
        }
        mpd_finishCommand (trConn);
        ErrorCheck(28,"mpd_sendListCommand");

      if (count)
      {
        QListViewItem *itemPointer = dataBaseView->firstChild();
        itemPointer->setText(0,QString::number(count) + " albums");
      }
      UnlockStatUpdate();
    }

    else
    if (searchMode == 2) //// Title ////
    {
    LockStatUpdate();
    dataMode = DMODE_DIR;
    qm_ListItem * ItemToAdd;
    mpd_InfoEntity * ntity;
    mpd_sendSearchCommand(trConn, MPD_TABLE_TITLE, searchString.utf8());
    if (ErrorCheck(30,"mpd_getNextInfoEntity"))
    {
        while((ntity = mpd_getNextInfoEntity(trConn)) != NULL)
        {
            if(ntity->type == MPD_INFO_ENTITY_TYPE_SONG)
            {
                mpd_Song *newSong = mpd_songDup (ntity->info.song);
                // artist
                QString songName = QString::fromUtf8(ntity->info.song->artist);
                if (songName.isEmpty()) songName = "Artist?";
                // plus title
                QString temp = QString::fromUtf8(ntity->info.song->title);
                songName += (" - " + temp);
                ItemToAdd = new qm_ListItem(RoOt, *newSong, songName);
                ItemToAdd->setType(LITEM_TRACK);
                ItemToAdd->setDragEnabled(TRUE);
                ItemToAdd->setExpandable(FALSE);
                ItemToAdd->setPixmap(0,trackIcon);
                count++;
            }
            mpd_freeInfoEntity(ntity);
        }
    }
    mpd_finishCommand(trConn);
    ErrorCheck(30,"mpd_getNextInfoEntity");

    if (count)
    {
        QListViewItem *itemPointer = dataBaseView->firstChild();
        itemPointer->setText(0,QString::number(count) + " tracks");
    }
        UnlockStatUpdate();
    }

    else
    if (searchMode == 3) //// Genre ////
    {
        LockStatUpdate();
    dataMode = DMODE_ALBUM;
    qm_ListItem * ItemToAdd;
    mpd_InfoEntity * ntity;
    mpd_startSearch(trConn, 1); // 1 is Exact. 0 is Not Exact.
    ErrorCheck(31,"mpd_startSearch");
    mpd_addConstraintSearch(trConn, MPD_TAG_ITEM_GENRE, searchString.utf8());
        mpd_commitSearch(trConn);
    while((ntity = mpd_getNextInfoEntity(trConn)) != NULL)
    {
        if(ntity->type == MPD_INFO_ENTITY_TYPE_SONG)
        {
            mpd_Song *newSong = mpd_songDup (ntity->info.song);
            // artist
            QString songName = QString::fromUtf8(ntity->info.song->artist);
            if (songName.isEmpty()) songName = "Artist?";
            // plus title
            QString temp = QString::fromUtf8(ntity->info.song->title);
            songName += (" - " + temp);
            ItemToAdd = new qm_ListItem(RoOt, *newSong, songName);
            ItemToAdd->setType(LITEM_TRACK);
            ItemToAdd->setDragEnabled(TRUE);
            ItemToAdd->setExpandable(FALSE);
            ItemToAdd->setPixmap(0,trackIcon);
            count++;
        }
        mpd_freeInfoEntity(ntity);
    }
    mpd_finishCommand(trConn);
    ErrorCheck(30,"mpd_commitSearch");

    if (count)
    {
        QListViewItem *itemPointer = dataBaseView->firstChild();
        itemPointer->setText(0,QString::number(count) + " tracks");
    }
    else // Nothing found; present a list of available genres
    {
        char *genreName;
        qm_ListItem * subItemToAdd;
        int genreCount = 0;

        dataBaseView->setSelectionMode(QListView::Single);

        ItemToAdd = new qm_ListItem(RoOt, "Available genres", LITEM_STAT_PLITEM); // PLITEM looks nice here
        ItemToAdd->setExpandable(TRUE);
            ItemToAdd->setSelectable(FALSE);

        mpd_startFieldSearch(trConn, MPD_TAG_ITEM_GENRE);
        mpd_commitSearch(trConn);
        while ( (genreName = mpd_getNextTag(trConn, MPD_TAG_ITEM_GENRE)) != NULL )
        {

            QString theGenre = QString::fromUtf8(genreName);

            subItemToAdd = new qm_ListItem(ItemToAdd, theGenre, LITEM_STAT_PLITEM);
            subItemToAdd->setType(LITEM_GENRE);
                subItemToAdd->setSelectable(TRUE);
            subItemToAdd->setExpandable(FALSE);
            subItemToAdd->setDragEnabled(FALSE);
            free(genreName);
            genreCount++;
        }
        mpd_finishCommand(trConn); // commitsearch
        ErrorCheck(31,"mpd_commitSearch");

        if (genreCount == 0)
        {
            subItemToAdd = new qm_ListItem(ItemToAdd, "No genres in database!", LITEM_STAT_MESSAGE);
                subItemToAdd->setSelectable(FALSE);
        }
        else
            dataMode = DMODE_GENRE; // make subitems clickable
    }
    UnlockStatUpdate();
    } // END IF Genre


    if(!count)
    {
        QListViewItem *itemPointer = dataBaseView->firstChild();
        itemPointer->setText(0,"0 matches");
        itemPointer->setSelectable(FALSE);
        itemPointer->setSelected(FALSE);
        itemPointer->setDragEnabled(FALSE);
        itemPointer->setPixmap(0,crossIcon);

            dataMode = 10; // invalid mode: disabled
    }

    b_noSearchNow = FALSE;
    RoOt->setOpen(TRUE);
}

// signal from the listview after a click
void qm_Tracks::dataListClicked(QListViewItem *theItem)
{
    playListView->clearFocus();
    dataBaseView->setFocus();

    // Clicked, but not on an item - or on a disabled item.
    if (!theItem || !theItem->isEnabled())
    return;

    // if it is the genre list: special treatment
    if ( ((qm_ListItem *)theItem)->getType() == LITEM_GENRE)
    {
    // put genre in search box
    le_searchstring->setText( theItem->text(0) );
    return;
    }

    // deselect up
    QListViewItem *itemUp = theItem->parent();
    while(itemUp)
    {
    if (itemUp->isSelected())
    {
        itemUp->setSelected(FALSE);
        dataBaseView->repaintItem(itemUp);
    }
    itemUp = itemUp->parent();
    }
    // deselect down
    unSelectSubs(theItem);
}

void qm_Tracks::dbvResized()
{
    QValueList<int> list = (*theSplitter).sizes();
    QValueList<int>::Iterator it = list.at(0);
    // *it is the child-widget's width
    dbasePanelWidth = *it;
    int ViewWidth = dataBaseView->viewportSize(*it, 0).width();
    dataBaseView->setColumnWidth(0, ViewWidth);

}

void qm_Tracks::plvResized()
{
    QValueList<int> list = (*theSplitter).sizes();
    QValueList<int>::Iterator it = list.at(1);
    // *it is the child-widget's width
    plistPanelWidth = *it;
    int ViewWidth = playListView->viewportSize(*it, 0).width();
    playListView->setColumnWidth(2, ViewWidth - playListView->columnWidth(0) - playListView->columnWidth(1) - playListView->columnWidth(3));
}

// Set the column width (signal from core)
void qm_Tracks::trSetSplitter(int dbViewW, int plViewW)
{
    // set the panels (can't use int, must be a list)
    QValueList<int> List;
    List.append(dbViewW);
    List.append(plViewW);
    (*theSplitter).setSizes(List);
    // set the column widths
    plvResized();
    dbvResized();
}

void qm_Tracks::updDbase()
{
    // dataBaseView->setEnabled(FALSE);
    cb_mode->setEnabled(FALSE);
    bt_gosearch->setEnabled(FALSE);
    if (b_colorise)
    {
    QPalette pal (dataBaseView->palette());
    pal.setColor(QColorGroup::Text, QColor(210, 70, 120));
    pal.setColor(QColorGroup::Button, QColor(240, 245, 250));
    bt_dbupd->setPalette(pal);
    }
    bt_dbupd->setEnabled(FALSE);
    bt_dbupd->setText("Busy...");
    b_noSearchNow = TRUE; // no searching during update
    crDBaseUpdate();
}


void qm_Tracks::trSetStatistics(int artists, int albums, int songs)
{
    totalTracks = songs;

    if (artists < 0)
    {
    lb_dbstats->setText("Statistics not available");
    }
    else
    {
    lb_dbstats->setText("Artists: " + QString::number(artists)
    + "   Albums: " + QString::number(albums)
    + "   Tracks: " + QString::number(songs));
    }


    // enable things after a databaseupdate
    if(trConn != NULL)
    {
    // dataBaseView->setEnabled(TRUE);
    cb_mode->setEnabled(TRUE);
    bt_gosearch->setEnabled(TRUE);
    bt_dbupd->setEnabled(TRUE);
        b_noSearchNow = FALSE;
    }

    if (b_colorise)
    {dataBaseView->setEnabled(TRUE);dataBaseView->setEnabled(TRUE);
    QPalette pal (dataBaseView->palette());
    pal.setColor(QColorGroup::Button, QColor(240, 245, 250));
    bt_dbupd->setPalette(pal);
    }
    bt_dbupd->setText("Update");
}
